/*----------------------------------------------------------------------------*
 * 
 *	file: spike-code.c
 * 
 *	function: test if lookup table might enhance speed in SLISSOM synapses 
 *	author:   Yoonsuck Choe <yschoe@cs.utexas.edu>
 *	creation: Tue Sep 23 03:27:24 CDT 1997
 * 
 *	$Revision: 1.3 $ $Date: 1999/02/09 07:42:25 $
 *	$Log: spike-code.cc,v $
 *	Revision 1.3  1999/02/09 07:42:25  yschoe
 *	init_normalized_leaky_table() added
 *
 *	Revision 1.2  1998/10/05 19:48:46  yschoe
 *	1. added get_inh|exc|affw_matrix() functions
 *	2. added setup_weight() functions
 *	3. separated calc_affresp and calc_latresp from calc_netinput
 *	4. set_matrix() for gaussian weight setup
 *
 *	Revision 1.1  1998/07/23 01:07:05  yschoe
 *	jul22 stable version (except netlayer)
 *
 *
 *----------------------------------------------------------------------------*/
#include "spike-code.h"
#include "real.h"

/* define MAX # of past spikes that can be stored: number of bits */
#define MAXQLEN	16

/* function prototypes */
static void 	errorexit(char *func, char*msg);
REAL 	*init_leaky_table(int bits, REAL decay_rate);
REAL 	*alloc_leaky_table(int bits);
REAL 	calc_table_entry(int bitpattern, int bits, REAL decay_rate);

/*----------------------------------------------------------------------------*
 *	errorexit() 
 *	  - display function where it was called and a message and exit
 *----------------------------------------------------------------------------*/
static void 	errorexit(char *func, char*msg)
{
  fprintf(stderr,"%s(): %s\n",func,msg);
  exit(1);
}

/*----------------------------------------------------------------------------*
 *	alloc_leaky_table() 
 *	  - malloc leaky table  : called by init_leaky_table()
 *----------------------------------------------------------------------------*/
REAL 	*alloc_leaky_table(int bits)
{
  REAL  *table;
  int   table_len;

  /* support storage of up to MAXQLEN past spikes */
  if (bits>MAXQLEN) 
  {
	errorexit("init_leaky_table","bits exceed MAXQLEN");
  }

  /* allocate memory for the lookup table */ 
  table_len = (int) (pow(2.0,bits));
  table = (REAL *) calloc(table_len,sizeof(REAL));

  return (table);
}
  

/*----------------------------------------------------------------------------*
 *	init_leaky_table() 
 *	  - malloc (call alloc_leaky_table()) and init leaky table values
 *----------------------------------------------------------------------------*/
REAL 	*init_leaky_table(int bits, REAL decay_rate)
{

  int	i;
  REAL  *table;
  int   table_len;

  /* allocate memory for the lookup table */ 
  table = alloc_leaky_table(bits);
  
  table_len = (int) pow(2.0,bits);

  /* initialize table values for look up */
  for (i=0; i<table_len; ++i) 
  {
	/* Using i bit by bit, calculate the decayed sum */
	/* Assume that we shift left when a spike comes in */
	/* thus bit 0 is the oldest slot */
  	table[i] = calc_table_entry(i,bits,decay_rate);
  }
  return (table);
}



/*----------------------------------------------------------------------------*
 *	init_normalized_leaky_table() 
 *	  - divide sum by number of bits
 *----------------------------------------------------------------------------*/
REAL 	*init_normalized_leaky_table(int bits, REAL decay_rate){

   int i, table_len;
   REAL *table;

   table = init_leaky_table(bits,decay_rate);

   table_len = (int) pow(2.0,bits);
   for (i=0; i<table_len; ++i)
   {
	table[i] /= (REAL)bits;
   }

   return table;

}

/*----------------------------------------------------------------------------*
 *	calc_table_entry() 
 *	  - return REAL value corresponding to the bitpattern
 *----------------------------------------------------------------------------*/
REAL 	calc_table_entry(int bitpattern, int bits, REAL decay_rate)
{

  REAL	value;
  int 	lsb; /* Least Significant Bit */
  int	k;
  int	bp;

  value = 0.0; bp = bitpattern;
# ifdef TESTDRIVE
  printf("[idx:%d] ",bitpattern);
# endif
  for (k=0; k<bits; ++k)
  { 
    /* get LSB */
    lsb = bp & 0x1;  
#   ifdef TESTDRIVE
    printf("%d ", lsb); 
#   endif
    value += (REAL) lsb / (REAL)pow(exp(decay_rate),k);   
    /* shift right */
    bp >>= 1;
  }
# ifdef TESTDRIVE
  printf("= %f\n",value);
# endif 
  return (value);
}


#ifdef TESTDRIVE
/*----------------------------------------------------------------------------*
 *
 *	main() 
 *	  - main test-driver
 *
 *----------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
  int	i;
  int	bits;
  int	table_len;
  REAL 	*table, lambda;

  /* process command line arguments */
  if (argc!=3)
  {
  	printf("spike-code (usage): spike-code bits lambda(decay-rate)\n");
	errorexit("main","command line incorrect.");
   }

  /* get command line arguments */
  bits	= atoi(argv[1]);
  sscanf(argv[2],"%f",&lambda);

  /* init table */
  table = init_leaky_table(bits,lambda);

  table_len = (int) pow(2.0,bits);

  /* print table */
  for (i=0; i<table_len; ++i)
  {
	printf("bits= %d, %d: %f \n",bits,i,table[i]);
  }
   
  return (0);
}
#endif
