//----------------------------------------------------------------------------
// 
//	file: spikelayer2-2.cc
// 
//	function: make spikelayer more robust to parameter change 
//		- 00jun6 : implement it with exact equations
//	author:   Yoonsuck Choe <yschoe@cs.utexas.edu>
//	creation: Tue Jun  6 18:41:10 CDT 2000
// 
//	$Revision: 1.1 $ $Date: 2001/11/30 20:24:15 $
//	$Log: spikelayer2-2.cc,v $
//	Revision 1.1  2001/11/30 20:24:15  yschoe
//	Initial revision
//
//
//	Revision: 1.3 Date: 2000/01/30 14:56:21 
//	Log: spikelayer2_2.cc,v 
//	Revision 1.3  2000/01/30 14:56:21  yschoe
//	*** empty log message ***
//
//	Revision 1.2  1999/12/22 05:29:48  yschoe
//	print_stat added
//
//	Revision 1.1  1999/02/09 07:42:25  yschoe
//	Initial revision
//
//
//----------------------------------------------------------------------------

#include "spikelayer2-2.h"

//----------------------------------------------------------------------------
//	spikelayer2_2() 
//	  - Constructor
//----------------------------------------------------------------------------
spikelayer2_2::spikelayer2_2(char *name, int cols, int rows, 
		 int exc_rad, int inh_rad,
                 int max_inp_layers, netlayer *feeding,
                 int inp_cols, int inp_rows, int  aff_rad,
                 REAL synapse_decay, REAL inhloop_decay, REAL theta, REAL tau,
		 int padded_flag)
	: spikelayer(name, cols, rows, exc_rad, inh_rad, max_inp_layers,
		 feeding, inp_cols, inp_rows, aff_rad, synapse_decay, 
		 inhloop_decay, theta, tau, padded_flag) {

  synapse_decay_mult = exp(-synapse_decay);

  min_net_input = 0.0;
  max_net_input = 1.0;
  // synapse_lut  = init_normalized_leaky_table(SPIKE_QLEN,synapse_decay);
  // inhloop_lut  = init_normalized_leaky_table(SPIKE_QLEN,inhloop_decay);

  auto_theta_flag = 0;  // by default
  auto_theta_rate = 0.475;  // by default

  avg_rate = 0.92;

  net_bb->add("lambda_exc_syn",&lambda_exc_syn);
  net_bb->add("lambda_inh_syn",&lambda_inh_syn);
  net_bb->add("lambda_int_syn",&lambda_int_syn);
  net_bb->add("lambda_loop",&lambda_loop);
  net_bb->add("avg_rate",&avg_rate);

}

//----------------------------------------------------------------------------
//	reset_neurons() 
//	  - reset neuron states while training
//----------------------------------------------------------------------------
void spikelayer2_2::reset_neurons(void) {
	reset_matrix(s_exc,cols,rows);
	reset_matrix(s_inh,cols,rows);
	reset_matrix(s_int,cols,rows);
	reset_matrix(eta,cols,rows);
	reset_matrix(thresh,cols,rows);
	reset_matrix(r,cols,rows);
	reset_matrix(net_aff_input,cols,rows);
	reset_matrix(net_exc_input,cols,rows);
	reset_matrix(net_inh_input,cols,rows);
	reset_matrix(v,cols,rows);
	reset_matrix(lfp,cols,rows);
	reset_matrix(avg_spiking,cols,rows);
	reset_matrix(output,cols,rows);
}

//----------------------------------------------------------------------------
//	activate() 
//	  - activate spike layer
//----------------------------------------------------------------------------
void spikelayer2_2::activate(int settling_iter) {

  // calc_aff tells you what settling stage you are in.
  // so, if it is 0, calculate the afferents.

  // Precondition
  //	none
  // Postcondition:
  //	output[][] must contain the firing rate values for use in the
  //	learn() function.

  REAL exc_decay = exp(-lambda_exc_syn);
  REAL inh_decay = exp(-lambda_inh_syn);
  REAL int_decay = exp(-lambda_int_syn);
  REAL loop_decay = exp(-lambda_loop);
  REAL net_eta = 0.0;
  
  // 0. Batch process aff and lateral Contributions
  // PRE: s_exc, s_inh, and s_int are all setup properly
  // if (settling_iter==0) { // fixed 00sep21 was (calc_aff)
  	calc_affresp();
  // }
  calc_latresp_exc(s_exc);
  calc_latresp_inh(s_inh);
  // POST: net_aff_input, net_exc, net_inh matrices are all setup

  int spike_sum = 0;
  for (int i=0; i<cols; ++i) {
	for (int j=0; j<rows; ++j) {


  		// 1. Calculate spike
		spikeact_q[i][j] = spike[i][j] = (int)Heav(eta[i][j]-thresh[i][j]);
		spike_sum += spike[i][j];

		REAL sp = spike[i][j];

  		// 2. Leaked Output
		new_s_exc[i][j] = s_exc[i][j]*exc_decay + sp;
		new_s_inh[i][j] = s_inh[i][j]*inh_decay + sp;
		new_s_int[i][j] = s_int[i][j]*int_decay + sp;

		// 3. eta
		net_input[i][j] = net_aff_input[i][j]  // include intra_columnar aff
			+ lat_exc_contri * net_exc_input[i][j]
			- lat_inh_contri * net_inh_input[i][j];
		// if (net_eta>0.1) {
	  	//		cout << "net_eta " << net_eta << endl;
		// }
		new_eta[i][j] = act_func(net_input[i][j]);

		new_eta[i][j] = (1.0-intrinsic_noise)*new_eta[i][j] 
				+ intrinsic_noise*drand48();

		// 4. thresh
		new_thresh[i][j] = theta 
			+ tau * ( thresh[i][j]*loop_decay + sp);

		// 5. add refractory period
		if (hard_refractory>0.0) 
			new_thresh[i][j] += Heav(r[i][j])*1000.0;

		// 6. Membrane voltage
		v[i][j] = eta[i][j] - thresh[i][j] - theta
			+ (REAL)spike[i][j];

		if (v[i][j]<-2.0) v[i][j] = -2.0;

		// 7. Refractory period
		if (hard_refractory>0.0) {
			new_r[i][j] = r[i][j] 
				+ hard_refractory * spike[i][j] - 1;
			new_r[i][j] *= Heav(new_r[i][j]);
		}
	}
  }

  // calculate auto theta
  if (settling_iter ==0 && auto_theta_flag == 1) {
	    nonzero_minmax(new_eta, &min_net_input, &max_net_input);
	    theta = max_net_input * auto_theta_rate;
	    cout << "new_theta = " << theta << endl;
	    cout << "rate = " << auto_theta_rate << endl;
  }

  // print percent firing : exclude first and second settling
  if (settling_iter > 1 && spike_sum>0) {
	cout 	<< name << ".spike_prob = " 
	 	<< (spike_sum/((REAL)(cols*rows))) << endl;
  } 

  // Update state
  for (int i=0; i<cols; ++i) {
	for (int j=0; j<rows; ++j) {
		s_exc[i][j] 	= new_s_exc[i][j];
		s_inh[i][j] 	= new_s_inh[i][j];
		s_int[i][j] 	= new_s_int[i][j];
		eta[i][j]   	= new_eta[i][j];
		thresh[i][j] 	= new_thresh[i][j];
		r[i][j] 	= new_r[i][j];
	
		// Book keeping stuff -- for display, etc.
		lfp[i][j]	= (v[i][j]+2.0)/3.5;
		avg_spiking[i][j] = avg_spiking[i][j]*avg_rate 
				+ (1.0-avg_rate)*spike[i][j];
        	output[i][j] = avg_spiking[i][j];
		// if (output[i][j]>0.1) {
		//		cout << "avg_spiking " << output[i][j]<< endl;
		//}
	}
  }

}

//----------------------------------------------------------------------------
//      print_stat() 
//        - print status of (col,row) neuron.
//----------------------------------------------------------------------------
void spikelayer2_2::print_stat(int col, int row) {

  cout  << net_aff_input[col][row] << " "
        << (lat_exc_contri*net_exc_input[col][row]) << " "
        << (lat_inh_contri*net_inh_input[col][row]) << " "
        << act_func(net_input[col][row]) << " "
        << s_exc[col][row] << " "
        << theta << " "
        << v[col][row] << " " 
	<< col << " " << row << " " << v[col][row] << " "
	<< s_exc[col][row] << endl;

}
//----------------------------------------------------------------------------
void spikelayer2_2::print_stat(int cur_inp, int inp_col, int inp_row) {

  int col = (int)(inp_col * cols / (REAL)feedings[cur_inp]->cols);
  int row = (int)(inp_row * rows / (REAL)feedings[cur_inp]->rows);

  cout  << net_aff_input[col][row] << " "
        << (lat_exc_contri*net_exc_input[col][row]) << " "
        << (lat_inh_contri*net_inh_input[col][row]) << " "
        << act_func(net_input[col][row]) << " "
        << s_exc[col][row] << " "
        << theta << " "
        << v[col][row] << " " << col << " " << row << " " << v[col][row] << " "
	<< s_exc[col][row] << endl;

}

/*----------------------------------------------------------------------------*
 *	randomize_eta() 
 *	  - initialize eta
 *----------------------------------------------------------------------------*/
void	spikelayer2_2::randomize_eta(void){

  // set_matrix(eta,cols,rows,init_eta_randomness);
  for (int i=0; i<cols; ++i) {
	for (int j=0; j<rows; ++j) {
		eta[i][j] = -drand48()*init_eta_randomness;
	}
  }
}
